@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -76,12 +77,11 @@ RgMrow_zc_mul_i(GEN x, GEN y, long c, long i)
       default: s = gadd(s, gmulgs(gcoeff(x,i,j), t)); break;
     }
   }
-  if (!s) { avma = av; return gen_0; }
-  return gerepileupto(av, s);
+  return s? gerepileupto(av, s): gc_const(av, gen_0);
 }
 GEN
 RgMrow_zc_mul(GEN x, GEN y, long i) { return RgMrow_zc_mul_i(x,y,lg(y),i); }
-/* x non-empty t_MAT, y a compatible zc (dimension > 0). */
+/* x nonempty t_MAT, y a compatible zc (dimension > 0). */
 static GEN
 RgM_zc_mul_i(GEN x, GEN y, long c, long l)
 {
@@ -104,6 +104,40 @@ RgM_zm_mul(GEN x, GEN y)
   return z;
 }
 
+/* x[i,]*y, l = lg(y) > 1 */
+static GEN
+RgMrow_ZC_mul_i(GEN x, GEN y, long i, long l)
+{
+  pari_sp av = avma;
+  GEN t = gmul(gcoeff(x,i,1), gel(y,1)); /* l > 1 ! */
+  long j;
+  for (j=2; j<l; j++)
+    if (signe(gel(y,j))) t = gadd(t, gmul(gcoeff(x,i,j), gel(y,j)));
+  return gerepileupto(av,t);
+}
+
+/* compatible t_MAT * t_COL, lx = lg(x) = lg(y) > 1, l = lgcols(x) */
+static GEN
+RgM_ZC_mul_i(GEN x, GEN y, long lx, long l)
+{
+  GEN z = cgetg(l,t_COL);
+  long i;
+  for (i=1; i<l; i++) gel(z,i) = RgMrow_ZC_mul_i(x,y,i,lx);
+  return z;
+}
+
+/* mostly useful when y is sparse */
+GEN
+RgM_ZM_mul(GEN x, GEN y)
+{
+  long j, c, l = lg(x), ly = lg(y);
+  GEN z = cgetg(ly, t_MAT);
+  if (l == 1) return z;
+  c = lgcols(x);
+  for (j = 1; j < ly; j++) gel(z,j) = RgM_ZC_mul_i(x, gel(y,j), l,c);
+  return z;
+}
+
 static GEN
 RgV_zc_mul_i(GEN x, GEN y, long l)
 {
@@ -349,7 +383,6 @@ Rg_RgC_sub(GEN a, GEN x)
   return z;
 }
 
-
 static GEN
 RgC_add_i(GEN x, GEN y, long lx)
 {
@@ -468,6 +501,52 @@ GEN
 RgMrow_RgC_mul(GEN x, GEN y, long i)
 { return RgMrow_RgC_mul_i(x, y, i, lg(x)); }
 
+static GEN
+RgM_RgC_mul_FpM(GEN x, GEN y, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    r = Flc_to_ZC_inplace(Flm_Flc_mul(RgM_to_Flm(x, pp),
+                                  RgV_to_Flv(y, pp), pp));
+  }
+  else
+    r = FpM_FpC_mul(RgM_to_FpM(x, p), RgC_to_FpC(y, p), p);
+  return gerepileupto(av, FpC_to_mod(r, p));
+}
+
+static GEN
+RgM_RgC_mul_FqM(GEN x, GEN y, GEN pol, GEN p)
+{
+  pari_sp av = avma;
+  GEN b, T = RgX_to_FpX(pol, p);
+  if (signe(T) == 0) pari_err_OP("*", x, y);
+  b = FqM_FqC_mul(RgM_to_FqM(x, T, p), RgC_to_FqC(y, T, p), T, p);
+  return gerepileupto(av, FqC_to_mod(b, T, p));
+}
+
+#define code(t1,t2) ((t1 << 6) | t2)
+static GEN
+RgM_RgC_mul_fast(GEN x, GEN y)
+{
+  GEN p, pol;
+  long pa;
+  long t = RgM_RgC_type(x,y, &p,&pol,&pa);
+  switch(t)
+  {
+    case t_INT:    return ZM_ZC_mul(x,y);
+    case t_FRAC:   return QM_QC_mul(x,y);
+    case t_FFELT:  return FFM_FFC_mul(x, y, pol);
+    case t_INTMOD: return RgM_RgC_mul_FpM(x, y, p);
+    case code(t_POLMOD, t_INTMOD):
+                   return RgM_RgC_mul_FqM(x, y, pol, p);
+    default:       return NULL;
+  }
+}
+#undef code
+
 /* compatible t_MAT * t_COL, lx = lg(x) = lg(y) > 1, l = lgcols(x) */
 static GEN
 RgM_RgC_mul_i(GEN x, GEN y, long lx, long l)
@@ -482,14 +561,11 @@ GEN
 RgM_RgC_mul(GEN x, GEN y)
 {
   long lx = lg(x);
-  GEN ffx = NULL, ffy = NULL;
+  GEN z;
   if (lx != lg(y)) pari_err_OP("operation 'RgM_RgC_mul'", x,y);
   if (lx == 1) return cgetg(1,t_COL);
-  if (RgM_is_FFM(x, &ffx) && RgC_is_FFC(y, &ffy)) {
-    if (!FF_samefield(ffx, ffy))
-      pari_err_OP("*", ffx, ffy);
-    return FFM_FFC_mul(x, y, ffx);
-  }
+  z = RgM_RgC_mul_fast(x, y);
+  if (z) return z;
   return RgM_RgC_mul_i(x, y, lx, lgcols(x));
 }
 
@@ -604,7 +680,7 @@ RgM_sqr_fast(GEN x)
   switch(t)
   {
     case t_INT:    return ZM_sqr(x);
-    case t_FRAC:   return QM_mul(x, x);
+    case t_FRAC:   return QM_sqr(x);
     case t_FFELT:  return FFM_mul(x, x, pol);
     case t_INTMOD: return RgM_mul_FpM(x, x, p);
     case code(t_POLMOD, t_INT):
@@ -897,7 +973,7 @@ RgM_isscalar(GEN x, GEN s)
     for (i=1; i<j; )
       if (!gequal0(gel(c,i++))) return 0;
     /* i = j */
-      if (!gequal(gel(c,i++),s)) return 0;
+    if (!gequal(gel(c,i++),s)) return 0;
     for (   ; i<lx; )
       if (!gequal0(gel(c,i++))) return 0;
   }
@@ -917,7 +993,7 @@ RgM_isidentity(GEN x)
     for (i=1; i<j; )
       if (!gequal0(gel(c,i++))) return 0;
     /* i = j */
-      if (!gequal1(gel(c,i++))) return 0;
+    if (!gequal1(gel(c,i++))) return 0;
     for (   ; i<lx; )
       if (!gequal0(gel(c,i++))) return 0;
   }
@@ -956,20 +1032,7 @@ RgM_isdiagonal(GEN x)
   return 1;
 }
 int
-isdiagonal(GEN x)
-{
-  return (typ(x)==t_MAT) && RgM_isdiagonal(x);
-}
-
-/* returns the first index i<=n such that x=v[i] if it exists, 0 otherwise */
-long
-RgV_isin(GEN v, GEN x)
-{
-  long i, l = lg(v);
-  for (i = 1; i < l; i++)
-    if (gequal(gel(v,i), x)) return i;
-  return 0;
-}
+isdiagonal(GEN x) { return (typ(x)==t_MAT) && RgM_isdiagonal(x); }
 
 GEN
 RgM_det_triangular(GEN mat)
