@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -153,6 +154,51 @@ RgXV_RgV_eval(GEN Q, GEN x)
   return v;
 }
 
+GEN
+RgX_homogenous_evalpow(GEN P, GEN A, GEN B)
+{
+  pari_sp av = avma;
+  long d, i, v;
+  GEN s;
+  if (typ(P)!=t_POL)
+    return mkvec2(P, gen_1);
+  d = degpol(P); v = varn(A);
+  s = scalarpol_shallow(gel(P, d+2), v);
+  for (i = d-1; i >= 0; i--)
+  {
+    s = gadd(gmul(s, A), gmul(gel(B,d+1-i), gel(P,i+2)));
+    if (gc_needed(av,1))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"RgX_homogenous_eval(%ld)",i);
+      s = gerepileupto(av, s);
+    }
+  }
+  s = gerepileupto(av, s);
+  return mkvec2(s, gel(B,d+1));
+}
+
+GEN
+QXQX_homogenous_evalpow(GEN P, GEN A, GEN B, GEN T)
+{
+  pari_sp av = avma;
+  long i, d = degpol(P), v = varn(A);
+  GEN s;
+  if (signe(P)==0) return mkvec2(pol_0(v), pol_1(v));
+  s = scalarpol_shallow(gel(P, d+2), v);
+  for (i = d-1; i >= 0; i--)
+  {
+    GEN c = gel(P,i+2), b = gel(B,d+1-i);
+    s = RgX_add(QXQX_mul(s, A, T), typ(c)==t_POL ? QXQX_QXQ_mul(b, c, T): gmul(b, c));
+    if (gc_needed(av,1))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"QXQX_homogenous_eval(%ld)",i);
+      s = gerepileupto(av, s);
+    }
+  }
+  s = gerepileupto(av, s);
+  return mkvec2(s, gel(B,d+1));
+}
+
 const struct bb_algebra *
 get_Rg_algebra(void)
 {
@@ -223,7 +269,7 @@ Rg_get_0(GEN x)
   switch(tx)
   {
     case t_INTMOD: retmkintmod(gen_0, icopy(p));
-    case t_PADIC: return cvtop(gen_0, p, lx);
+    case t_PADIC: return zeropadic(p, lx);
     case t_FFELT: return FF_zero(T);
     default: return gen_0;
   }
@@ -393,6 +439,34 @@ RgX_deflate(GEN x0, long d)
   return y;
 }
 
+/* F a t_RFRAC */
+long
+rfrac_deflate_order(GEN F)
+{
+  GEN N = gel(F,1), D = gel(F,2);
+  long m = (degpol(D) <= 0)? 0: RgX_deflate_order(D);
+  if (m == 1) return 1;
+  if (typ(N) == t_POL && varn(N) == varn(D))
+    m = cgcd(m, RgX_deflate_order(N));
+  return m;
+}
+/* F a t_RFRAC */
+GEN
+rfrac_deflate_max(GEN F, long *m)
+{
+  *m = rfrac_deflate_order(F);
+  return rfrac_deflate(F, *m);
+}
+/* F a t_RFRAC */
+GEN
+rfrac_deflate(GEN F, long m)
+{
+  GEN N = gel(F,1), D = gel(F,2);
+  if (m == 1) return F;
+  if (typ(N) == t_POL && varn(N) == varn(D)) N = RgX_deflate(N, m);
+  D = RgX_deflate(D, m); return mkrfrac(N, D);
+}
+
 /* return x0(X^d) */
 GEN
 RgX_inflate(GEN x0, long d)
@@ -409,37 +483,37 @@ RgX_inflate(GEN x0, long d)
 }
 
 /* return P(X + c) using destructive Horner, optimize for c = 1,-1 */
-GEN
-RgX_translate(GEN P, GEN c)
+static GEN
+RgX_translate_basecase(GEN P, GEN c)
 {
   pari_sp av = avma;
-  GEN Q, *R;
+  GEN Q, R;
   long i, k, n;
 
   if (!signe(P) || gequal0(c)) return RgX_copy(P);
   Q = leafcopy(P);
-  R = (GEN*)(Q+2); n = degpol(P);
-  if (equali1(c))
+  R = Q+2; n = degpol(P);
+  if (isint1(c))
   {
     for (i=1; i<=n; i++)
     {
-      for (k=n-i; k<n; k++) R[k] = gadd(R[k], R[k+1]);
+      for (k=n-i; k<n; k++) gel(R,k) = gadd(gel(R,k), gel(R,k+1));
       if (gc_needed(av,2))
       {
-        if(DEBUGMEM>1) pari_warn(warnmem,"TR_POL(1), i = %ld/%ld", i,n);
-        Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
+        if(DEBUGMEM>1) pari_warn(warnmem,"RgX_translate(1), i = %ld/%ld", i,n);
+        Q = gerepilecopy(av, Q); R = Q+2;
       }
     }
   }
-  else if (equalim1(c))
+  else if (isintm1(c))
   {
     for (i=1; i<=n; i++)
     {
-      for (k=n-i; k<n; k++) R[k] = gsub(R[k], R[k+1]);
+      for (k=n-i; k<n; k++) gel(R,k) = gsub(gel(R,k), gel(R,k+1));
       if (gc_needed(av,2))
       {
-        if(DEBUGMEM>1) pari_warn(warnmem,"TR_POL(-1), i = %ld/%ld", i,n);
-        Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
+        if(DEBUGMEM>1) pari_warn(warnmem,"RgX_translate(-1), i = %ld/%ld", i,n);
+        Q = gerepilecopy(av, Q); R = Q+2;
       }
     }
   }
@@ -447,88 +521,56 @@ RgX_translate(GEN P, GEN c)
   {
     for (i=1; i<=n; i++)
     {
-      for (k=n-i; k<n; k++) R[k] = gadd(R[k], gmul(c, R[k+1]));
+      for (k=n-i; k<n; k++) gel(R,k) = gadd(gel(R,k), gmul(c, gel(R,k+1)));
       if (gc_needed(av,2))
       {
-        if(DEBUGMEM>1) pari_warn(warnmem,"TR_POL, i = %ld/%ld", i,n);
-        Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
+        if(DEBUGMEM>1) pari_warn(warnmem,"RgX_translate, i = %ld/%ld", i,n);
+        Q = gerepilecopy(av, Q); R = Q+2;
       }
     }
   }
   return gerepilecopy(av, Q);
 }
-
-/* return P(X + c) using destructive Horner, optimize for c = 1,-1 */
 GEN
-ZX_translate(GEN P, GEN c)
+RgX_translate(GEN P, GEN c)
 {
   pari_sp av = avma;
-  GEN Q, *R;
-  long i, k, n;
-
-  if (!signe(P) || !signe(c)) return ZX_copy(P);
-  Q = leafcopy(P);
-  R = (GEN*)(Q+2); n = degpol(P);
-  if (equali1(c))
-  {
-    for (i=1; i<=n; i++)
-    {
-      for (k=n-i; k<n; k++) R[k] = addii(R[k], R[k+1]);
-      if (gc_needed(av,2))
-      {
-        if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate(1), i = %ld/%ld", i,n);
-        Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
-      }
-    }
-  }
-  else if (equalim1(c))
-  {
-    for (i=1; i<=n; i++)
-    {
-      for (k=n-i; k<n; k++) R[k] = subii(R[k], R[k+1]);
-      if (gc_needed(av,2))
-      {
-        if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate(-1), i = %ld/%ld", i,n);
-        Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
-      }
-    }
-  }
+  long n = degpol(P);
+  if (n < 40)
+    return RgX_translate_basecase(P, c);
   else
   {
-    for (i=1; i<=n; i++)
-    {
-      for (k=n-i; k<n; k++) R[k] = addmulii_inplace(R[k], c, R[k+1]);
-      if (gc_needed(av,2))
-      {
-        if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate, i = %ld/%ld", i,n);
-        Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
-      }
-    }
+    long d = n >> 1;
+    GEN Q = RgX_translate(RgX_shift_shallow(P, -d), c);
+    GEN R = RgX_translate(RgXn_red_shallow(P, d), c);
+    GEN S = gpowgs(deg1pol_shallow(gen_1, c, varn(P)), d);
+    return gerepileupto(av, RgX_add(RgX_mul(Q, S), R));
   }
-  return gerepilecopy(av, Q);
 }
+
 /* return lift( P(X + c) ) using Horner, c in R[y]/(T) */
 GEN
 RgXQX_translate(GEN P, GEN c, GEN T)
 {
   pari_sp av = avma;
-  GEN Q, *R;
+  GEN Q, R;
   long i, k, n;
 
   if (!signe(P) || gequal0(c)) return RgX_copy(P);
   Q = leafcopy(P);
-  R = (GEN*)(Q+2); n = degpol(P);
+  R = Q+2; n = degpol(P);
   for (i=1; i<=n; i++)
   {
     for (k=n-i; k<n; k++)
     {
       pari_sp av2 = avma;
-      R[k] = gerepileupto(av2, RgX_rem(gadd(R[k], gmul(c, R[k+1])), T));
+      gel(R,k) = gerepileupto(av2,
+                   RgX_rem(gadd(gel(R,k), gmul(c, gel(R,k+1))), T));
     }
     if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgXQX_translate, i = %ld/%ld", i,n);
-      Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
+      Q = gerepilecopy(av, Q); R = Q+2;
     }
   }
   return gerepilecopy(av, Q);
@@ -543,7 +585,7 @@ RgXQX_translate(GEN P, GEN c, GEN T)
 /* to INT / FRAC / (POLMOD mod T), not memory clean because T not copied,
  * but everything else is */
 static GEN
-QXQ_to_mod_copy(GEN x, GEN T)
+QXQ_to_mod(GEN x, GEN T)
 {
   long d;
   switch(typ(x))
@@ -560,8 +602,8 @@ QXQ_to_mod_copy(GEN x, GEN T)
   }
 }
 /* pure shallow version */
-static GEN
-QXQ_to_mod(GEN x, GEN T)
+GEN
+QXQ_to_mod_shallow(GEN x, GEN T)
 {
   long d;
   switch(typ(x))
@@ -577,14 +619,14 @@ QXQ_to_mod(GEN x, GEN T)
              return NULL;/* LCOV_EXCL_LINE */
   }
 }
-/* T a ZX, z lifted from (Q[Y]/(T(Y)))[X], apply QXQ_to_mod_copy to all coeffs.
+/* T a ZX, z lifted from (Q[Y]/(T(Y)))[X], apply QXQ_to_mod to all coeffs.
  * Not memory clean because T not copied, but everything else is */
 static GEN
 QXQX_to_mod(GEN z, GEN T)
 {
   long i,l = lg(z);
   GEN x = cgetg(l,t_POL);
-  for (i=2; i<l; i++) gel(x,i) = QXQ_to_mod_copy(gel(z,i), T);
+  for (i=2; i<l; i++) gel(x,i) = QXQ_to_mod(gel(z,i), T);
   x[1] = z[1]; return normalizepol_lg(x,l);
 }
 /* pure shallow version */
@@ -593,7 +635,7 @@ QXQX_to_mod_shallow(GEN z, GEN T)
 {
   long i,l = lg(z);
   GEN x = cgetg(l,t_POL);
-  for (i=2; i<l; i++) gel(x,i) = QXQ_to_mod(gel(z,i), T);
+  for (i=2; i<l; i++) gel(x,i) = QXQ_to_mod_shallow(gel(z,i), T);
   x[1] = z[1]; return normalizepol_lg(x,l);
 }
 /* Apply QXQX_to_mod to all entries. Memory-clean ! */
@@ -605,23 +647,23 @@ QXQXV_to_mod(GEN V, GEN T)
   for (i=1;i<l; i++) gel(z,i) = QXQX_to_mod(gel(V,i), T);
   return z;
 }
-/* Apply QXQ_to_mod_copy to all entries. Memory-clean ! */
+/* Apply QXQ_to_mod to all entries. Memory-clean ! */
 GEN
 QXQV_to_mod(GEN V, GEN T)
 {
   long i, l = lg(V);
   GEN z = cgetg(l, t_VEC); T = ZX_copy(T);
-  for (i=1;i<l; i++) gel(z,i) = QXQ_to_mod_copy(gel(V,i), T);
+  for (i=1;i<l; i++) gel(z,i) = QXQ_to_mod(gel(V,i), T);
   return z;
 }
 
-/* Apply QXQ_to_mod_copy to all entries. Memory-clean ! */
+/* Apply QXQ_to_mod to all entries. Memory-clean ! */
 GEN
 QXQC_to_mod_shallow(GEN V, GEN T)
 {
   long i, l = lg(V);
   GEN z = cgetg(l, t_COL);
-  for (i=1;i<l; i++) gel(z,i) = QXQ_to_mod(gel(V,i), T);
+  for (i=1;i<l; i++) gel(z,i) = QXQ_to_mod_shallow(gel(V,i), T);
   return z;
 }
 
@@ -871,6 +913,15 @@ RgX_mulXn(GEN x, long d)
   return gerepileupto(av, z);
 }
 
+long
+RgXV_maxdegree(GEN x)
+{
+  long d = -1, i, l = lg(x);
+  for (i = 1; i < l; i++)
+    d = maxss(d, degpol(gel(x,i)));
+  return d;
+}
+
 long
 RgX_val(GEN x)
 {
@@ -878,7 +929,7 @@ RgX_val(GEN x)
   if (lx == 2) return LONG_MAX;
   for (i = 2; i < lx; i++)
     if (!isexactzero(gel(x,i))) break;
-  if (i == lx) return LONG_MAX;/* possible with non-rational zeros */
+  if (i == lx) return LONG_MAX;/* possible with nonrational zeros */
   return i - 2;
 }
 long
@@ -888,7 +939,7 @@ RgX_valrem(GEN x, GEN *Z)
   if (lx == 2) { *Z = pol_0(varn(x)); return LONG_MAX; }
   for (i = 2; i < lx; i++)
     if (!isexactzero(gel(x,i))) break;
-  /* possible with non-rational zeros */
+  /* possible with nonrational zeros */
   if (i == lx) { *Z = pol_0(varn(x)); return LONG_MAX; }
   v = i - 2;
   *Z = RgX_shift_shallow(x, -v);
@@ -1536,18 +1587,33 @@ RgX_sqr_i(GEN x)
 /*******************************************************************/
 GEN
 RgX_Rg_divexact(GEN x, GEN y) {
-  long i, lx;
+  long i, lx = lg(x);
   GEN z;
-  if (typ(y) == t_INT && is_pm1(y))
-    return signe(y) < 0 ? RgX_neg(x): RgX_copy(x);
-  z = cgetg_copy(x, &lx); z[1] = x[1];
+  if (lx == 2) return gcopy(x);
+  switch(typ(y))
+  {
+    case t_INT:
+      if (is_pm1(y)) return signe(y) < 0 ? RgX_neg(x): RgX_copy(x);
+      break;
+    case t_INTMOD: case t_POLMOD: return RgX_Rg_mul(x, ginv(y));
+  }
+  z = cgetg(lx, t_POL); z[1] = x[1];
   for (i=2; i<lx; i++) gel(z,i) = gdivexact(gel(x,i),y);
   return z;
 }
 GEN
 RgX_Rg_div(GEN x, GEN y) {
-  long i, lx;
-  GEN z = cgetg_copy(x, &lx); z[1] = x[1];
+  long i, lx = lg(x);
+  GEN z;
+  if (lx == 2) return gcopy(x);
+  switch(typ(y))
+  {
+    case t_INT:
+      if (is_pm1(y)) return signe(y) < 0 ? RgX_neg(x): RgX_copy(x);
+      break;
+    case t_INTMOD: case t_POLMOD: return RgX_Rg_mul(x, ginv(y));
+  }
+  z = cgetg(lx, t_POL); z[1] = x[1];
   for (i=2; i<lx; i++) gel(z,i) = gdiv(gel(x,i),y);
   return normalizepol_lg(z, lx);
 }
@@ -1574,7 +1640,14 @@ GEN
 RgX_div_by_X_x(GEN a, GEN x, GEN *r)
 {
   long l = lg(a), i;
-  GEN a0, z0, z = cgetg(l-1, t_POL);
+  GEN a0, z0, z;
+
+  if (l <= 3)
+  {
+    if (r) *r = l == 2? gen_0: gcopy(gel(a,2));
+    return pol_0(0);
+  }
+  z = cgetg(l-1, t_POL);
   z[1] = a[1];
   a0 = a + l-1;
   z0 = z + l-2; *z0 = *a0--;
@@ -1636,7 +1709,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
   if (RgX_is_FpX(x, &p) && RgX_is_FpX(y, &p) && p)
   {
     z = FpX_divrem(RgX_to_FpX(x, p), RgX_to_FpX(y, p), p, pr);
-    if (!z) { avma = av; return NULL; }
+    if (!z) return gc_NULL(av);
     z = FpX_to_mod(z, p);
     if (!pr || pr == ONLY_REM || pr == ONLY_DIVIDES)
       return gerepileupto(av, z);
@@ -1669,7 +1742,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
     if (dx < dy) /* leading coeff of x was in fact zero */
     {
       if (pr == ONLY_DIVIDES) {
-        avma = av;
+        set_avma(av);
         return (dx < 0)? pol_0(varn(x)) : NULL;
       }
       if (pr == ONLY_REM)
@@ -1679,7 +1752,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
         else
         {
           GEN t;
-          avma = av;
+          set_avma(av);
           t = cgetg(dx + 3, t_POL); t[1] = x[1];
           for (i = 2; i < dx + 3; i++) gel(t,i) = gcopy(gel(x,i));
           return t;
@@ -1690,7 +1763,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
         if (dx < 0)
         {
           p2 = gclone(p2);
-          avma = av;
+          set_avma(av);
           z = pol_0(varn(x));
           x = scalarpol(p2, varn(x));
           gunclone(p2);
@@ -1698,7 +1771,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
         else
         {
           GEN t;
-          avma = av;
+          set_avma(av);
           z = pol_0(varn(x));
           t = cgetg(dx + 3, t_POL); t[1] = x[1];
           for (i = 2; i < dx + 3; i++) gel(t,i) = gcopy(gel(x,i));
@@ -1708,7 +1781,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
       }
       else
       {
-        avma = av;
+        set_avma(av);
         z = pol_0(varn(x));
       }
       return z;
@@ -1729,7 +1802,7 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
     for (j=i-dy+1; j<=i && j<=dz; j++) p1 = gsub(p1, gmul(gel(z,j),gel(y,i-j)));
     if (y_lead) p1 = simplify(f(p1,y_lead));
 
-    if (isrationalzero(p1)) { avma=av1; p1 = gen_0; }
+    if (isrationalzero(p1)) { set_avma(av1); p1 = gen_0; }
     else
       p1 = avma==av1? gcopy(p1): gerepileupto(av1,p1);
     gel(z,i-dy) = p1;
@@ -1743,19 +1816,18 @@ RgX_divrem_i(GEN x, GEN y, GEN *pr)
     /* we always enter this loop at least once */
     for (j=0; j<=i && j<=dz; j++) p1 = gsub(p1, gmul(gel(z,j),gel(y,i-j)));
     if (mod && avma==av1) p1 = gmul(p1,mod);
-    if (!gequal0(p1)) { sx = 1; break; } /* remainder is non-zero */
+    if (!gequal0(p1)) { sx = 1; break; } /* remainder is nonzero */
     if (!isexactzero(p1)) break;
     if (!i) break;
-    avma=av1;
+    set_avma(av1);
   }
   if (pr == ONLY_DIVIDES)
   {
-    if (sx) { avma=av; return NULL; }
-    avma = (pari_sp)rem;
-    return gerepileupto(av,z-2);
+    if (sx) return gc_NULL(av);
+    set_avma((pari_sp)rem); return gerepileupto(av,z-2);
   }
   lr=i+3; rem -= lr;
-  if (avma==av1) { avma = (pari_sp)rem; p1 = gcopy(p1); }
+  if (avma==av1) { set_avma((pari_sp)rem); p1 = gcopy(p1); }
   else p1 = gerepileupto((pari_sp)rem,p1);
   rem[0] = evaltyp(t_POL) | evallg(lr);
   rem[1] = z[-1];
@@ -1802,7 +1874,7 @@ RgXQX_divrem(GEN x, GEN y, GEN T, GEN *pr)
     if (pr)
     {
       av0 = avma; x = RgXQX_red(x, T);
-      if (pr == ONLY_DIVIDES) { avma=av0; return signe(x)? NULL: gen_0; }
+      if (pr == ONLY_DIVIDES) { set_avma(av0); return signe(x)? NULL: gen_0; }
       if (pr == ONLY_REM) return x;
       *pr = x;
     }
@@ -1822,7 +1894,7 @@ RgXQX_divrem(GEN x, GEN y, GEN T, GEN *pr)
   }
   av0 = avma; dz = dx-dy;
   lead = gequal1(lead)? NULL: gclone(ginvmod(lead,T));
-  avma = av0;
+  set_avma(av0);
   z = cgetg(dz+3,t_POL); z[1] = x[1];
   x += 2; y += 2; z += 2;
 
@@ -1835,7 +1907,7 @@ RgXQX_divrem(GEN x, GEN y, GEN T, GEN *pr)
     if (lead) p1 = gmul(grem(p1, T), lead);
     tetpil=avma; gel(z,i-dy) = gerepile(av,tetpil, grem(p1, T));
   }
-  if (!pr) { if (lead) gunclone(lead); return z-2; }
+  if (!pr) { guncloneNULL(lead); return z-2; }
 
   rem = (GEN)avma; av = (pari_sp)new_chunk(dx+3);
   for (sx=0; ; i--)
@@ -1844,13 +1916,13 @@ RgXQX_divrem(GEN x, GEN y, GEN T, GEN *pr)
     for (j=0; j<=i && j<=dz; j++) p1 = gsub(p1, gmul(gel(z,j),gel(y,i-j)));
     tetpil=avma; p1 = grem(p1, T); if (!gequal0(p1)) { sx = 1; break; }
     if (!i) break;
-    avma=av;
+    set_avma(av);
   }
   if (pr == ONLY_DIVIDES)
   {
-    if (lead) gunclone(lead);
-    if (sx) { avma=av0; return NULL; }
-    avma = (pari_sp)rem; return z-2;
+    guncloneNULL(lead);
+    if (sx) return gc_NULL(av0);
+    return gc_const((pari_sp)rem, z-2);
   }
   lr=i+3; rem -= lr;
   rem[0] = evaltyp(t_POL) | evallg(lr);
@@ -1865,7 +1937,7 @@ RgXQX_divrem(GEN x, GEN y, GEN T, GEN *pr)
     tetpil=avma; gel(rem,i) = gerepile(av,tetpil, grem(p1, T));
   }
   rem -= 2;
-  if (lead) gunclone(lead);
+  guncloneNULL(lead);
   if (!sx) (void)normalizepol_lg(rem, lr);
   if (pr == ONLY_REM) return gerepileupto(av0,rem);
   *pr = rem; return z-2;
@@ -1939,7 +2011,7 @@ ZXQX_dvd(GEN x, GEN y, GEN T)
       gerepilecoeffs(av2,x,dx+1);
     }
   }
-  avma = av; return (dx < 0);
+  return gc_bool(av, dx < 0);
 }
 
 /* T either NULL or a t_POL. */
@@ -2227,15 +2299,21 @@ GEN
 RgXn_powu_i(GEN x, ulong m, long n)
 {
   struct modXn S;
+  long v;
+  if (n == 1) return x;
+  v = RgX_valrem(x, &x);
+  if (v) n -= m * v;
   S.v = varn(x); S.n = n;
-  return gen_powu_i(x, m, (void*)&S,_sqrXn,_mulXn);
+  x = gen_powu_i(x, m, (void*)&S,_sqrXn,_mulXn);
+  if (v) x = RgX_shift_shallow(x, m * v);
+  return x;
 }
 GEN
 RgXn_powu(GEN x, ulong m, long n)
 {
-  struct modXn S;
-  S.v = varn(x); S.n = n;
-  return gen_powu(x, m, (void*)&S,_sqrXn,_mulXn);
+  pari_sp av;
+  if (n == 1) return gcopy(x);
+  av = avma; return gerepilecopy(av, RgXn_powu_i(x, m, n));
 }
 
 GEN
@@ -2338,7 +2416,10 @@ RgXn_inv_FpX(GEN x, long e, GEN p)
   if (lgefint(p) == 3)
   {
     ulong pp = uel(p, 2);
-    r = Flx_to_ZX_inplace(Flxn_inv(RgX_to_Flx(x, pp), e, pp));
+    if (pp == 2)
+      r = F2x_to_ZX(F2xn_inv(RgX_to_F2x(x), e));
+    else
+      r = Flx_to_ZX_inplace(Flxn_inv(RgX_to_Flx(x, pp), e, pp));
   }
   else
     r = FpXn_inv(RgX_to_FpX(x, p), e, p);
@@ -2352,7 +2433,7 @@ RgXn_inv_FpXQX(GEN x, long n, GEN pol, GEN p)
   GEN r, T = RgX_to_FpX(pol, p);
   if (signe(T) == 0) pari_err_OP("/", gen_1, x);
   r = FpXQXn_inv(RgX_to_FpXQX(x, T, p), n, T, p);
-  return gerepileupto(av, FpXQX_to_mod(r, pol, p));
+  return gerepileupto(av, FpXQX_to_mod(r, T, p));
 }
 
 #define code(t1,t2) ((t1 << 6) | t2)
@@ -2381,8 +2462,21 @@ RgXn_inv(GEN f, long e)
   return RgXn_inv_i(f, e);
 }
 
+/* Compute intformal(x^n*S)/x^(n+1) */
+static GEN
+RgX_integXn(GEN x, long n)
+{
+  long i, lx = lg(x);
+  GEN y;
+  if (lx == 2) return RgX_copy(x);
+  y = cgetg(lx, t_POL); y[1] = x[1];
+  for (i=2; i<lx; i++)
+    gel(y,i) = gdivgs(gel(x,i), n+i-1);
+  return RgX_renormalize_lg(y, lx);;
+}
+
 GEN
-RgXn_exp(GEN h, long e)
+RgXn_expint(GEN h, long e)
 {
   pari_sp av = avma, av2;
   long v = varn(h), n;
@@ -2393,27 +2487,38 @@ RgXn_exp(GEN h, long e)
   g = pol_1(v);
   n = 1; mask = quadratic_prec_mask(e);
   av2 = avma;
-  if (degpol(h) < 1 || !gequal0(gel(h,2)))
-    pari_err_DOMAIN("RgXn_exp","valuation", "<", gen_1, h);
   for (;mask>1;)
   {
-    GEN q, w;
+    GEN u, w;
     long n2 = n;
     n<<=1; if (mask & 1) n--;
     mask >>= 1;
-    g = RgX_sub(RgX_muls(g,2),RgXn_mul(f,RgXn_sqr(g,n2),n2));
-    q = RgX_deriv(RgXn_red_shallow(h,n2));
-    w = RgX_add(q, RgXn_mul(g, RgX_sub(RgX_deriv(f), RgXn_mul(f,q,n-1)),n-1));
-    f = RgX_add(f, RgXn_mul(f, RgX_sub(RgXn_red_shallow(h, n), RgX_integ(w)), n));
+    u = RgXn_mul(g, RgX_mulhigh_i(f, RgXn_red_shallow(h, n2-1), n2-1), n-n2);
+    u = RgX_add(u, RgX_shift_shallow(RgXn_red_shallow(h, n-1), 1-n2));
+    w = RgXn_mul(f, RgX_integXn(u, n2-1), n-n2);
+    f = RgX_add(f, RgX_shift_shallow(w, n2));
+    if (mask<=1) break;
+    u = RgXn_mul(g, RgXn_mulhigh(f, g, n2, n), n-n2);
+    g = RgX_sub(g, RgX_shift_shallow(u, n2));
     if (gc_needed(av2,2))
     {
-      if(DEBUGMEM>1) pari_warn(warnmem,"RgXn_exp, e = %ld", n);
+      if (DEBUGMEM>1) pari_warn(warnmem,"RgXn_expint, e = %ld", n);
       gerepileall(av2, 2, &f, &g);
     }
   }
   return gerepileupto(av, f);
 }
 
+GEN
+RgXn_exp(GEN h, long e)
+{
+  long d = degpol(h);
+  if (d < 0) return pol_1(varn(h));
+  if (!d || !gequal0(gel(h,2)))
+    pari_err_DOMAIN("RgXn_exp","valuation", "<", gen_1, h);
+  return RgXn_expint(RgX_deriv(h), e);
+}
+
 GEN
 RgXn_reverse(GEN f, long e)
 {
@@ -2493,14 +2598,12 @@ RgXn_sqrt(GEN h, long e)
 GEN
 RgXQ_powu(GEN x, ulong n, GEN T)
 {
-  pari_sp av;
-  GEN y;
+  pari_sp av = avma;
 
   if (!n) return pol_1(varn(x));
   if (n == 1) return RgX_copy(x);
-  av = avma;
-  y = gen_powu(x, n, (void*)T, &_sqr, &_mul);
-  return gerepileupto(av, y);
+  x = gen_powu_i(x, n, (void*)T, &_sqr, &_mul);
+  return gerepilecopy(av, x);
 }
 /* x,T in Rg[X], n in N, compute lift(x^n mod T)) */
 GEN
@@ -2508,15 +2611,38 @@ RgXQ_pow(GEN x, GEN n, GEN T)
 {
   pari_sp av;
   long s = signe(n);
-  GEN y;
 
   if (!s) return pol_1(varn(x));
   if (is_pm1(n) == 1)
     return (s < 0)? RgXQ_inv(x, T): RgX_copy(x);
   av = avma;
   if (s < 0) x = RgXQ_inv(x, T);
-  y = gen_pow(x, n, (void*)T, &_sqr, &_mul);
-  return gerepileupto(av, y);
+  x = gen_pow_i(x, n, (void*)T, &_sqr, &_mul);
+  return gerepilecopy(av, x);
+}
+static GEN
+_ZXQsqr(void *data, GEN x) { return ZXQ_sqr(x, (GEN)data); }
+static GEN
+_ZXQmul(void *data, GEN x, GEN y) { return ZXQ_mul(x,y, (GEN)data); }
+
+/* generates the list of powers of x of degree 0,1,2,...,l*/
+GEN
+ZXQ_powers(GEN x, long l, GEN T)
+{
+  int use_sqr = 2*degpol(x) >= degpol(T);
+  return gen_powers(x, l, use_sqr, (void *)T,_ZXQsqr,_ZXQmul,_one);
+}
+
+/* x,T in Z[X], n in N, compute lift(x^n mod T)) */
+GEN
+ZXQ_powu(GEN x, ulong n, GEN T)
+{
+  pari_sp av = avma;
+
+  if (!n) return pol_1(varn(x));
+  if (n == 1) return ZX_copy(x);
+  x = gen_powu_i(x, n, (void*)T, &_ZXQsqr, &_ZXQmul);
+  return gerepilecopy(av, x);
 }
 
 /* generates the list of powers of x of degree 0,1,2,...,l*/
@@ -2715,8 +2841,7 @@ RgX_mul_FpX(GEN x, GEN y, GEN p)
   }
   else
     r = FpX_mul(RgX_to_FpX(x, p), RgX_to_FpX(y, p), p);
-  if (signe(r)==0)
-  { avma = av; return zero_FpX_mod(p, varn(x)); }
+  if (signe(r)==0) { set_avma(av); return zero_FpX_mod(p, varn(x)); }
   return gerepileupto(av, FpX_to_mod(r, p));
 }
 
@@ -2741,8 +2866,7 @@ RgX_mul_FpXQX(GEN x, GEN y, GEN pol, GEN p)
   kx = ZXX_to_Kronecker(RgX_to_FpXQX(x, T, p), dT);
   ky = ZXX_to_Kronecker(RgX_to_FpXQX(y, T, p), dT);
   r = FpX_mul(kx, ky, p);
-  if (signe(r)==0)
-  { avma = av; return zero_FpXQX_mod(pol, p, varn(x)); }
+  if (signe(r)==0) { set_avma(av); return zero_FpXQX_mod(pol, p, varn(x)); }
   return gerepileupto(av, Kronecker_to_mod(FpX_to_mod(r, p), pol));
 }
 
@@ -2772,8 +2896,7 @@ RgX_sqr_FpX(GEN x, GEN p)
   }
   else
     r = FpX_sqr(RgX_to_FpX(x, p), p);
-  if (signe(r)==0)
-  { avma = av; return zero_FpX_mod(p, varn(x)); }
+  if (signe(r)==0) { set_avma(av); return zero_FpX_mod(p, varn(x)); }
   return gerepileupto(av, FpX_to_mod(r, p));
 }
 
@@ -2787,8 +2910,7 @@ RgX_sqr_FpXQX(GEN x, GEN pol, GEN p)
   dT = degpol(T);
   kx = ZXX_to_Kronecker(RgX_to_FpXQX(x, T, p), dT);
   r = FpX_sqr(kx, p);
-  if (signe(r)==0)
-  { avma = av; return zero_FpXQX_mod(pol, p, varn(x)); }
+  if (signe(r)==0) { set_avma(av); return zero_FpXQX_mod(pol, p, varn(x)); }
   return gerepileupto(av, Kronecker_to_mod(FpX_to_mod(r, p), pol));
 }
 
@@ -2814,8 +2936,7 @@ RgX_rem_FpX(GEN x, GEN y, GEN p)
   }
   else
     r = FpX_rem(RgX_to_FpX(x, p), RgX_to_FpX(y, p), p);
-  if (signe(r)==0)
-  { avma = av; return zero_FpX_mod(p, varn(x)); }
+  if (signe(r)==0) { set_avma(av); return zero_FpX_mod(p, varn(x)); }
   return gerepileupto(av, FpX_to_mod(r, p));
 }
 
@@ -2843,8 +2964,7 @@ RgX_rem_FpXQX(GEN x, GEN y, GEN pol, GEN p)
   }
   else
     r = FpXQX_rem(RgX_to_FpXQX(x, T, p), RgX_to_FpXQX(y, T, p), T, p);
-  if (signe(r)==0)
-  { avma = av; return zero_FpXQX_mod(pol, p, varn(x)); }
+  if (signe(r)==0) { set_avma(av); return zero_FpXQX_mod(pol, p, varn(x)); }
   return gerepileupto(av, FpXQX_to_mod(r, T, p));
 }
 
