@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -19,10 +20,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /*                                                                 */
 /*******************************************************************/
 static GEN
-RgX_to_ser_i(GEN x, long l, long lx, long v, int copy)
+RgX_to_ser_i(GEN x, long l, long v, int copy)
 {
+  long i, lx = lg(x);
   GEN y;
-  long i;
   if (lx == 2) return zeroser(varn(x), l-2);
   if (l < 2) pari_err_BUG("RgX_to_ser (l < 2)");
   y = cgetg(l,t_SER); y[1] = x[1];
@@ -38,7 +39,7 @@ RgX_to_ser_i(GEN x, long l, long lx, long v, int copy)
 }
 /* enlarge/truncate t_POL x to a t_SER with lg l */
 GEN
-RgX_to_ser(GEN x, long l) { return RgX_to_ser_i(x, l, lg(x), RgX_val(x), 0); }
+RgX_to_ser(GEN x, long l) { return RgX_to_ser_i(x, l, RgX_val(x), 0); }
 GEN
 RgX_to_ser_inexact(GEN x, long l)
 {
@@ -50,7 +51,7 @@ RgX_to_ser_inexact(GEN x, long l)
       pari_warn(warner,"normalizing a series with 0 leading term");
       first = 0;
     }
-  return RgX_to_ser_i(x, l, lx, i - 2, 0);
+  return RgX_to_ser_i(x, l, i - 2, 0);
 }
 GEN
 rfrac_to_ser(GEN x, long l)
@@ -96,25 +97,29 @@ sertoser(GEN x, long prec)
   return y;
 }
 
+/* R(1/x) = x^v * n/d, val(n) = val(d) = 0 */
+long
+rfracrecip(GEN *pn, GEN *pd)
+{
+  long v = degpol(*pd);
+  if (typ(*pn) == t_POL && varn(*pn) == varn(*pd))
+  {
+    v -= degpol(*pn);
+    (void)RgX_valrem(*pn, pn); *pn = RgX_recip(*pn);
+  }
+  (void)RgX_valrem(*pd, pd); *pd = RgX_recip(*pd);
+  return v;
+}
+
 /* R(1/x) + O(x^N) */
 GEN
 rfracrecip_to_ser_absolute(GEN R, long N)
 {
   GEN n = gel(R,1), d = gel(R,2);
-  long vx = varn(d), vn, v, dn;
-
-  if (typ(n) != t_POL || varn(n) != vx) { vn = 0; dn = 0; }
-  else
-  {
-    vn = RgX_valrem(n, &n);
-    n = RgX_recip(n);
-    dn = degpol(n);
-  }
-  v = vn - RgX_valrem(d, &d);
-  d = RgX_recip(d);
-  R = gdiv(n, RgX_to_ser(d, N+2));
-  setvalp(R, valp(R) + degpol(d)-dn-v);
-  return R;
+  long v = rfracrecip(&n, &d); /* R(1/x) = x^v * n/d, val(n) = val(d) = 0 */
+  if (N <= v) return zeroser(varn(d), N);
+  R = gdiv(n, RgX_to_ser(d, N-v+2));
+  setvalp(R, v); return R;
 }
 
 /* assume prec >= 0 */
@@ -157,7 +162,7 @@ poltoser(GEN x, long v, long prec)
   long s = varncmp(varn(x), v);
   if (s < 0) err_ser_priority(x,v);
   if (s > 0) return scalarser(x, v, prec);
-  return RgX_to_ser_i(x, prec+2, lg(x), RgX_val(x), 1);
+  return RgX_to_ser_i(x, prec+2, RgX_val(x), 1);
 }
 /* x a t_RFRAC */
 static GEN
