@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -61,7 +62,7 @@ idealtyp(GEN *ideal, GEN *arch)
   {
     case t_MAT: lx = lg(x);
       if (lx == 1) { t = id_PRINCIPAL; x = gen_0; break; }
-      if (lx != lgcols(x)) pari_err_TYPE("idealtyp [non-square t_MAT]",x);
+      if (lx != lgcols(x)) pari_err_TYPE("idealtyp [nonsquare t_MAT]",x);
       t = id_MAT;
       break;
 
@@ -133,7 +134,7 @@ idealhnf_shallow(GEN nf, GEN x)
 {
   long tx = typ(x), lx = lg(x), N;
 
-  /* cannot use idealtyp because here we allow non-square matrices */
+  /* cannot use idealtyp because here we allow nonsquare matrices */
   if (tx == t_VEC && lx == 3) { x = gel(x,1); tx = typ(x); lx = lg(x); }
   if (tx == t_VEC && lx == 6) return pr_hnf(nf,x); /* PRIME */
   switch(tx)
@@ -161,7 +162,7 @@ idealhnf_shallow(GEN nf, GEN x)
       GEN A = gel(x,1), B = gel(x,2);
       N = nf_get_degree(nf);
       if (N != 2)
-        pari_err_TYPE("idealhnf [Qfb for non-quadratic fields]", x);
+        pari_err_TYPE("idealhnf [Qfb for nonquadratic fields]", x);
       if (!equalii(qfb_disc(x), D))
         pari_err_DOMAIN("idealhnf [Qfb]", "disc(q)", "!=", D, x);
       /* x -> A Z + (-B + sqrt(D)) / 2 Z
@@ -309,8 +310,7 @@ static int
 ok_elt(GEN x, GEN xZ, GEN y)
 {
   pari_sp av = avma;
-  int r = ZM_equal(x, ZM_hnfmodid(y, xZ));
-  avma = av; return r;
+  return gc_bool(av, ZM_equal(x, ZM_hnfmodid(y, xZ)));
 }
 
 static GEN
@@ -362,7 +362,7 @@ get_random_a(GEN nf, GEN x, GEN xZ)
   setlg(mul, lm);
   setlg(beta,lm);
   z = cgetg(lm, t_VECSMALL);
-  for(av = avma;; avma = av)
+  for(av = avma;; set_avma(av))
   {
     for (a=NULL,i=1; i<lm; i++)
     {
@@ -504,7 +504,7 @@ idealHNF_Z_factor_i(GEN x, GEN f0, GEN *pvN, GEN *pvZ)
   for (i = 1; i < l; i++)
   {
     GEN p = gel(P,i);
-    vZ[i] = f0? Z_pval(xZ, p): itou(gel(E,i));
+    vZ[i] = f0? Z_pval(xZ, p): (long) itou(gel(E,i));
     vN[i] = idealHNF_norm_pval(x,p, vZ[i]);
   }
   return P;
@@ -612,7 +612,7 @@ idealHNF_factor_i(GEN nf, GEN x, GEN cx, GEN FA)
       GEN P = gel(L,j);
       pari_sp av = avma;
       v = idealHNF_val(x, P, Nval, Zval);
-      avma = av;
+      set_avma(av);
       Nval -= v*pr_get_f(P);
       v += vc * pr_get_e(P); if (!v) continue;
       gel(vP,k) = P;
@@ -652,11 +652,16 @@ idealHNF_factor(GEN nf, GEN x, ulong lim)
   if (lim)
   {
     GEN P, E;
-    long l;
-    F = Z_factor_limit(gcoeff(x,1,1), lim);
-    P = gel(F,1); l = lg(P);
+    long i;
+    /* strict useless because of prime table */
+    F = absZ_factor_limit(gcoeff(x,1,1), lim);
+    P = gel(F,1);
     E = gel(F,2);
-    if (l > 1 && abscmpiu(gel(P,l-1), lim) >= 0) { setlg(P,l-1); setlg(E,l-1); }
+    /* filter out entries > lim */
+    for (i = lg(P)-1; i; i--)
+      if (cmpiu(gel(P,i), lim) <= 0) break;
+    setlg(P, i+1);
+    setlg(E, i+1);
   }
   x = Q_primitive_part(x, &cx);
   return idealHNF_factor_i(nf, x, cx, F);
@@ -686,13 +691,11 @@ Q_nffactor(GEN nf, GEN y, ulong lim)
   else
   {
     f = Q_factor_limit(y, lim);
-    P = gel(f,1); l = lg(P);
+    P = gel(f,1);
     E = gel(f,2);
-    for (i = l-1; i > 0; i--)
-    {
+    for (i = lg(P)-1; i > 0; i--)
       if (abscmpiu(gel(P,i), lim) < 0) break;
-      setlg(P,i); setlg(E,i);
-    }
+    setlg(P,i+1); setlg(E,i+1);
   }
   P = gel(f,1); l = lg(P); if (l == 1) return f;
   E = gel(f,2);
@@ -707,6 +710,32 @@ Q_nffactor(GEN nf, GEN y, ulong lim)
   gel(f,2) = E; return f;
 }
 
+GEN
+idealfactor_partial(GEN nf, GEN x, GEN L)
+{
+  pari_sp av = avma;
+  long i, j, l;
+  GEN P, E;
+  if (!L) return idealfactor(nf, x);
+  if (typ(L) == t_INT) return idealfactor_limit(nf, x, itou(L));
+  l = lg(L); if (l == 1) return trivial_fact();
+  P = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+  {
+    GEN p = gel(L,i);
+    gel(P,i) = typ(p) == t_INT? idealprimedec(nf, p): mkvec(p);
+  }
+  P = shallowconcat1(P); settyp(P, t_COL);
+  P = gen_sort_uniq(P, (void*)&cmp_prime_ideal, &cmp_nodata);
+  E = cgetg_copy(P, &l);
+  for (i = j = 1; i < l; i++)
+  {
+    long v = idealval(nf, x, gel(P,i));
+    if (v) { gel(P,j) = gel(P,i); gel(E,j) = stoi(v); j++; }
+  }
+  setlg(P,j);
+  setlg(E,j); return gerepilecopy(av, mkmat2(P, E));
+}
 GEN
 idealfactor_limit(GEN nf, GEN x, ulong lim)
 {
@@ -714,12 +743,12 @@ idealfactor_limit(GEN nf, GEN x, ulong lim)
   GEN fa, y;
   long tx = idealtyp(&x,&y);
 
-  nf = checknf(nf);
   if (tx == id_PRIME)
   {
     if (lim && abscmpiu(pr_get_p(x), lim) >= 0) return trivial_fact();
     retmkmat2(mkcolcopy(x), mkcol(gen_1));
   }
+  nf = checknf(nf);
   if (tx == id_PRINCIPAL)
   {
     y = nf_to_scalar_or_basis(nf, x);
@@ -839,19 +868,18 @@ idealispower(GEN nf, GEN A, long n, GEN *pB)
   if (n <= 0) pari_err_DOMAIN("idealispower", "n", "<=", gen_0, stoi(n));
   if (n == 1) { if (pB) *pB = idealhnf(nf,A); return 1; }
   v = idealnumden(nf,A);
-  if (gequal0(gel(v,1))) { avma = av; if (pB) *pB = cgetg(1,t_MAT); return 1; }
+  if (gequal0(gel(v,1))) { set_avma(av); if (pB) *pB = cgetg(1,t_MAT); return 1; }
   if (!idealsqrtn_int(nf, gel(v,1), n, pB? &N: NULL)) return 0;
   if (!idealsqrtn_int(nf, gel(v,2), n, pB? &D: NULL)) return 0;
-  if (pB) *pB = gerepileupto(av, idealdiv(nf,N,D)); else avma = av;
+  if (pB) *pB = gerepileupto(av, idealdiv(nf,N,D)); else set_avma(av);
   return 1;
 }
 
-/* x t_INT or integral non-0 ideal in HNF */
+/* x t_INT or integral nonzero ideal in HNF */
 static GEN
 idealredmodpower_i(GEN nf, GEN x, ulong k, ulong B)
 {
   GEN cx, y, U, N, F, Q;
-  long nF;
   if (typ(x) == t_INT)
   {
     if (!signe(x) || is_pm1(x)) return gen_1;
@@ -860,16 +888,13 @@ idealredmodpower_i(GEN nf, GEN x, ulong k, ulong B)
     return ginv(factorback(F));
   }
   N = gcoeff(x,1,1); if (is_pm1(N)) return gen_1;
-  F = Z_factor_limit(N, B); nF=lg(gel(F,1))-1;
-  if (BPSW_psp(gcoeff(F,nF,1))) U = NULL;
-  else
+  F = absZ_factor_limit_strict(N, B, &U);
+  if (U)
   {
-    GEN M = powii(gcoeff(F,nF,1), gcoeff(F,nF,2));
+    GEN M = powii(gel(U,1), gel(U,2));
     y = hnfmodid(x, M); /* coprime part to B! */
     if (!idealispower(nf, y, k, &U)) U = NULL;
     x = hnfmodid(x, diviiexact(N, M));
-    setlg(gel(F,1), nF); /* remove last entry (unfactored part) */
-    setlg(gel(F,2), nF);
   }
   /* x = B-smooth part of initial x */
   x = Q_primitive_part(x, &cx);
@@ -890,7 +915,7 @@ idealredmodpower(GEN nf, GEN x, ulong n, ulong B)
   if (!n) pari_err_DOMAIN("idealredmodpower","n", "=", gen_0, gen_0);
   x = idealnumden(nf, x);
   a = gel(x,1);
-  if (isintzero(a)) { avma = av; return gen_1; }
+  if (isintzero(a)) { set_avma(av); return gen_1; }
   a = idealredmodpower_i(nf, gel(x,1), n, B);
   b = idealredmodpower_i(nf, gel(x,2), n, B);
   if (!isint1(b)) a = nf_to_scalar_or_basis(nf, nfdiv(nf, a, b));
@@ -913,7 +938,7 @@ idealval(GEN nf, GEN A, GEN P)
   A = Q_primitive_part(A, &cA);
   p = pr_get_p(P);
   vcA = cA? Q_pval(cA,p): 0;
-  if (pr_is_inert(P)) { avma = av; return vcA; }
+  if (pr_is_inert(P)) return gc_long(av,vcA);
   Zval = Z_pval(gcoeff(A,1,1), p);
   if (!Zval) v = 0;
   else
@@ -921,7 +946,7 @@ idealval(GEN nf, GEN A, GEN P)
     long Nval = idealHNF_norm_pval(A, p, Zval);
     v = idealHNF_val(A, P, Nval, Zval);
   }
-  avma = av; return vcA? v + vcA*pr_get_e(P): v;
+  return gc_long(av, vcA? v + vcA*pr_get_e(P): v);
 }
 GEN
 gpidealval(GEN nf, GEN ix, GEN P)
@@ -955,7 +980,7 @@ idealadd(GEN nf, GEN x, GEN y)
   if (is_pm1(a))
   {
     long N = lg(x)-1;
-    if (!dz) { avma = av; return matid(N); }
+    if (!dz) { set_avma(av); return matid(N); }
     return gerepileupto(av, scalarmat(ginv(dz), N));
   }
   z = ZM_hnfmodid(shallowconcat(x,y), a);
@@ -1026,7 +1051,7 @@ idealaddmultoone(GEN nf, GEN list)
   L = cgetg(l, t_VEC);
   if (l == 1)
     pari_err_DOMAIN("idealaddmultoone", "sum(ideals)", "!=", gen_1, L);
-  nz = 0; /* number of non-zero ideals in L */
+  nz = 0; /* number of nonzero ideals in L */
   for (i=1; i<l; i++)
   {
     GEN I = gel(list,i);
@@ -1168,6 +1193,23 @@ famat_add(GEN f, GEN x)
   }
   return h;
 }
+/* add x^-1 to famat f */
+static GEN
+famat_sub(GEN f, GEN x)
+{
+  GEN h = cgetg(3,t_MAT);
+  if (lgcols(f) == 1)
+  {
+    gel(h,1) = mkcolcopy(x);
+    gel(h,2) = mkcol(gen_m1);
+  }
+  else
+  {
+    gel(h,1) = append(gel(f,1), x);
+    gel(h,2) = gconcat(gel(f,2), gen_m1);
+  }
+  return h;
+}
 
 GEN
 famat_mul(GEN f, GEN g)
@@ -1188,6 +1230,25 @@ famat_mul(GEN f, GEN g)
   return h;
 }
 
+GEN
+famat_div(GEN f, GEN g)
+{
+  GEN h;
+  if (typ(g) != t_MAT) {
+    if (typ(f) == t_MAT) return famat_sub(f, g);
+    h = cgetg(3, t_MAT);
+    gel(h,1) = mkcol2(gcopy(f), gcopy(g));
+    gel(h,2) = mkcol2(gen_1, gen_m1);
+  }
+  if (typ(f) != t_MAT) return famat_sub(g, f);
+  if (lgcols(f) == 1) return famat_inv(g);
+  if (lgcols(g) == 1) return gcopy(f);
+  h = cgetg(3,t_MAT);
+  gel(h,1) = gconcat(gel(f,1), gel(g,1));
+  gel(h,2) = gconcat(gel(f,2), gneg(gel(g,2)));
+  return h;
+}
+
 GEN
 famat_sqr(GEN f)
 {
@@ -1292,13 +1353,23 @@ famat_to_nf(GEN nf, GEN f)
   return t;
 }
 
+GEN
+famat_idealfactor(GEN nf, GEN x)
+{
+  long i, l;
+  GEN g = gel(x,1), e = gel(x,2), h = cgetg_copy(g, &l);
+  for (i = 1; i < l; i++) gel(h,i) = idealfactor(nf, gel(g,i));
+  h = famat_reduce(famatV_factorback(h,e));
+  return sort_factor(h, (void*)&cmp_prime_ideal, &cmp_nodata);
+}
+
 GEN
 famat_reduce(GEN fa)
 {
   GEN E, G, L, g, e;
   long i, k, l;
 
-  if (lgcols(fa) == 1) return fa;
+  if (typ(fa) != t_MAT || lgcols(fa) == 1) return fa;
   g = gel(fa,1); l = lg(g);
   e = gel(fa,2);
   L = gen_indexsort(g, (void*)&cmp_universal, &cmp_nodata);
@@ -1326,6 +1397,30 @@ famat_reduce(GEN fa)
   setlg(G, k);
   setlg(E, k); return mkmat2(G,E);
 }
+GEN
+matreduce(GEN f)
+{ pari_sp av = avma;
+  switch(typ(f))
+  {
+    case t_VEC: case t_COL:
+    {
+      GEN e; f = vec_reduce(f, &e);
+      return gerepilecopy(av, mkmat2(f, zc_to_ZC(e)));
+    }
+    case t_MAT:
+      if (lg(f) == 3) break;
+    default:
+      pari_err_TYPE("matreduce", f);
+  }
+  if (typ(gel(f,1)) == t_VECSMALL)
+    f = famatsmall_reduce(f);
+  else
+  {
+    if (!RgV_is_ZV(gel(f,2))) pari_err_TYPE("matreduce",f);
+    f = famat_reduce(f);
+  }
+  return gerepilecopy(av, f);
+}
 
 GEN
 famatsmall_reduce(GEN fa)
@@ -1361,6 +1456,40 @@ famatsmall_reduce(GEN fa)
   setlg(E, k); return mkmat2(G,E);
 }
 
+GEN
+famat_remove_trivial(GEN fa)
+{
+  GEN P, E, p = gel(fa,1), e = gel(fa,2);
+  long j, k, l = lg(p);
+  P = cgetg(l, t_COL);
+  E = cgetg(l, t_COL);
+  for (j = k = 1; j < l; j++)
+    if (signe(gel(e,j))) { gel(P,k) = gel(p,j); gel(E,k++) = gel(e,j); }
+  setlg(P, k); setlg(E, k); return mkmat2(P,E);
+}
+
+GEN
+famatV_factorback(GEN v, GEN e)
+{
+  long i, l = lg(e);
+  GEN V;
+  if (l == 1) return trivial_fact();
+  V = signe(gel(e,1))? famat_pow_shallow(gel(v,1), gel(e,1)): trivial_fact();
+  for (i = 2; i < l; i++) V = famat_mulpow_shallow(V, gel(v,i), gel(e,i));
+  return V;
+}
+
+GEN
+famatV_zv_factorback(GEN v, GEN e)
+{
+  long i, l = lg(e);
+  GEN V;
+  if (l == 1) return trivial_fact();
+  V = uel(e,1)? famat_pows_shallow(gel(v,1), uel(e,1)): trivial_fact();
+  for (i = 2; i < l; i++) V = famat_mulpows_shallow(V, gel(v,i), uel(e,i));
+  return V;
+}
+
 GEN
 ZM_famat_limit(GEN fa, GEN limit)
 {
@@ -1536,6 +1665,29 @@ famat_makecoprime(GEN nf, GEN g, GEN e, GEN pr, GEN prk, GEN EX)
   return famat_to_nf_modideal_coprime(nf, G, E, prk, EX);
 }
 
+/* simplified version of famat_makecoprime for X = SUnits[1] */
+GEN
+sunits_makecoprime(GEN X, GEN pr, GEN prk)
+{
+  GEN G, p = pr_get_p(pr), prkZ = gcoeff(prk,1,1);
+  long i, l = lg(X);
+
+  G = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+  {
+    GEN x = gel(X,i);
+    if (typ(x) == t_INT) /* a prime */
+      x = equalii(x,p)? p_makecoprime(pr): modii(x, prkZ);
+    else
+    {
+      (void)ZC_nfvalrem(x, pr, &x); /* x *= (b/p)^v_pr(x) */
+      x = ZC_hnfrem(FpC_red(x,prkZ), prk);
+    }
+    gel(G,i) = x;
+  }
+  return G;
+}
+
 /* prod g[i]^e[i] mod bid, assume (g[i], id) = 1 and 1 < lg(g) <= lg(e) */
 GEN
 famat_to_nf_moddivisor(GEN nf, GEN g, GEN e, GEN bid)
@@ -1544,7 +1696,8 @@ famat_to_nf_moddivisor(GEN nf, GEN g, GEN e, GEN bid)
   if (lg(cyc) == 1)
     t = gen_1;
   else
-    t = famat_to_nf_modideal_coprime(nf, g, e, bid_get_ideal(bid), gel(cyc,1));
+    t = famat_to_nf_modideal_coprime(nf, g, e, bid_get_ideal(bid),
+                                     cyc_get_expo(cyc));
   return set_sign_mod_divisor(nf, mkmat2(g,e), t, bid_get_sarch(bid));
 }
 
@@ -1592,7 +1745,7 @@ idealmulelt(GEN nf, GEN x, GEN A)
   D = mulii(zkmultable_capZ(x), gcoeff(A,1,1));
   x = zkC_multable_mul(A, x);
   settyp(x, t_MAT); lx = lg(x);
-  /* x may contain scalars (at most 1 since the ideal is non-0)*/
+  /* x may contain scalars (at most 1 since the ideal is nonzero)*/
   for (i=1; i<lx; i++)
     if (typ(gel(x,i)) == t_INT)
     {
@@ -1726,8 +1879,8 @@ idealsqr_aux(GEN nf, GEN x, long tx)
         x = scalarmat(x, N);
       }
       else
-      {
-        x = ZM_hnfmodid(m, gcdii(sqri(a), zkmultable_capZ(m)));
+      { /* could use gcdii(sqri(a), zkmultable_capZ(m)), but costly */
+        x = ZM_hnfmodid(m, sqri(a));
         if (cx) cx = gsqr(cx);
         if (cx) x = ZM_Q_mul(x, cx);
       }
@@ -1753,7 +1906,7 @@ GEN
 idealnorm(GEN nf, GEN x)
 {
   pari_sp av;
-  GEN y, T;
+  GEN y;
   long tx;
 
   switch(idealtyp(&x,&y))
@@ -1762,15 +1915,114 @@ idealnorm(GEN nf, GEN x)
     case id_MAT: return RgM_det_triangular(x);
   }
   /* id_PRINCIPAL */
-  nf = checknf(nf); T = nf_get_pol(nf); av = avma;
-  x = nf_to_scalar_or_alg(nf, x);
-  x = (typ(x) == t_POL)? RgXQ_norm(x, T): gpowgs(x, degpol(T));
+  nf = checknf(nf); av = avma;
+  x = nfnorm(nf, x);
   tx = typ(x);
   if (tx == t_INT) return gerepileuptoint(av, absi(x));
   if (tx != t_FRAC) pari_err_TYPE("idealnorm",x);
   return gerepileupto(av, Q_abs(x));
 }
 
+/* x \cap Z */
+GEN
+idealdown(GEN nf, GEN x)
+{
+  pari_sp av = avma;
+  GEN y, c;
+  switch(idealtyp(&x,&y))
+  {
+    case id_PRIME: return icopy(pr_get_p(x));
+    case id_MAT: return gcopy(gcoeff(x,1,1));
+  }
+  /* id_PRINCIPAL */
+  nf = checknf(nf); av = avma;
+  x = nf_to_scalar_or_basis(nf, x);
+  if (is_rational_t(typ(x))) return Q_abs(x);
+  x = Q_primitive_part(x, &c);
+  y = zkmultable_capZ(zk_multable(nf, x));
+  return gerepilecopy(av, mul_content(c, y));
+}
+
+/* true nf */
+static GEN
+idealismaximal_int(GEN nf, GEN p)
+{
+  GEN L;
+  if (!BPSW_psp(p)) return NULL;
+  if (!dvdii(nf_get_index(nf), p) &&
+      !FpX_is_irred(FpX_red(nf_get_pol(nf),p), p)) return NULL;
+  L = idealprimedec(nf, p);
+  return lg(L) == 2? gel(L,1): NULL;
+}
+/* true nf */
+static GEN
+idealismaximal_mat(GEN nf, GEN x)
+{
+  GEN p, c, L;
+  long i, l, f;
+  x = Q_primitive_part(x, &c);
+  p = gcoeff(x,1,1);
+  if (c)
+  {
+    if (typ(c) == t_FRAC || !equali1(p)) return NULL;
+    return idealismaximal_int(nf, p);
+  }
+  if (!BPSW_psp(p)) return NULL;
+  l = lg(x); f = 1;
+  for (i = 2; i < l; i++)
+  {
+    c = gcoeff(x,i,i);
+    if (equalii(c, p)) f++; else if (!equali1(c)) return NULL;
+  }
+  L = idealprimedec_limit_f(nf, p, f);
+  for (i = lg(L)-1; i; i--)
+  {
+    GEN pr = gel(L,i);
+    if (pr_get_f(pr) != f) break;
+    if (idealval(nf, x, pr) == 1) return pr;
+  }
+  return NULL;
+}
+/* true nf */
+static GEN
+idealismaximal_i(GEN nf, GEN x)
+{
+  GEN L, p, pr, c;
+  long i, l;
+  switch(idealtyp(&x,&c))
+  {
+    case id_PRIME: return x;
+    case id_MAT: return idealismaximal_mat(nf, x);
+  }
+  /* id_PRINCIPAL */
+  nf = checknf(nf);
+  x = nf_to_scalar_or_basis(nf, x);
+  switch(typ(x))
+  {
+    case t_INT: return idealismaximal_int(nf, absi_shallow(x));
+    case t_FRAC: return NULL;
+  }
+  x = Q_primitive_part(x, &c);
+  if (c) return NULL;
+  p = zkmultable_capZ(zk_multable(nf, x));
+  L = idealprimedec(nf, p); l = lg(L); pr = NULL;
+  for (i = 1; i < l; i++)
+  {
+    long v = ZC_nfval(x, gel(L,i));
+    if (v > 1 || (v && pr)) return NULL;
+    pr = gel(L,i);
+  }
+  return pr;
+}
+GEN
+idealismaximal(GEN nf, GEN x)
+{
+  pari_sp av = avma;
+  x = idealismaximal_i(checknf(nf), x);
+  if (!x) { set_avma(av); return gen_0; }
+  return gerepilecopy(av, x);
+}
+
 /* I^(-1) = { x \in K, Tr(x D^(-1) I) \in Z }, D different of K/Q
  *
  * nf[5][6] = pp( D^(-1) ) = pp( HNF( T^(-1) ) ), T = (Tr(wi wj))
@@ -1958,7 +2210,7 @@ idealpowprime(GEN nf, GEN pr, GEN n, GEN *pc)
   return mkvec2(q, gen);
 }
 
-/* x * pr^n. Assume x in HNF or scalar (possibly non-integral) */
+/* x * pr^n. Assume x in HNF or scalar (possibly nonintegral) */
 GEN
 idealmulpowprime(GEN nf, GEN x, GEN pr, GEN n)
 {
@@ -2027,8 +2279,8 @@ idealpow_aux(GEN nf, GEN x, long tx, GEN n)
         x = scalarmat(x, N);
       }
       else
-      {
-        x = ZM_hnfmodid(m, gcdii(powii(a,n1), zkmultable_capZ(m)));
+      { /* could use gcdii(powii(a,n1), zkmultable_capZ(m)), but costly */
+        x = ZM_hnfmodid(m, powii(a,n1));
         if (cx) cx = powgi(cx,n);
         if (s<0) {
           GEN xZ = gcoeff(x,1,1);
@@ -2084,17 +2336,17 @@ _sqr(void *data, GEN x) { return _idealsqrred((GEN)data, x); }
 GEN
 idealpowred(GEN nf, GEN x, GEN n)
 {
-  pari_sp av = avma;
+  pari_sp av = avma, av2;
   long s;
   GEN y;
 
   if (typ(n) != t_INT) pari_err_TYPE("idealpowred",n);
   s = signe(n); if (s == 0) return idealpow(nf,x,n);
-  y = gen_pow(x, n, (void*)nf, &_sqr, &_mul);
-
+  y = gen_pow_i(x, n, (void*)nf, &_sqr, &_mul);
+  av2 = avma;
   if (s < 0) y = idealinv(nf,y);
   if (s < 0 || is_pm1(n)) y = idealred(nf,y);
-  return gerepileupto(av,y);
+  return avma == av2? gerepilecopy(av,y): gerepileupto(av,y);
 }
 
 GEN
@@ -2130,11 +2382,11 @@ isideal(GEN nf,GEN x)
   if (!ZM_ishnf(x)) return 0;
   xZ = gcoeff(x,1,1);
   for (j=2; j<=N; j++)
-    if (!dvdii(xZ, gcoeff(x,j,j))) { avma = av; return 0; }
+    if (!dvdii(xZ, gcoeff(x,j,j))) return gc_long(av,0);
   for (i=2; i<=N; i++)
     for (j=2; j<=N; j++)
-      if (! hnf_invimage(x, zk_ei_mul(nf,gel(x,i),j))) { avma = av; return 0; }
-  avma=av; return 1;
+       if (! hnf_invimage(x, zk_ei_mul(nf,gel(x,i),j))) return gc_long(av,0);
+  return gc_long(av,1);
 }
 
 GEN
@@ -2178,7 +2430,7 @@ idealdivexact(GEN nf, GEN x0, GEN y0)
   x = idealhnf_shallow(nf, x0);
   y = idealhnf_shallow(nf, y0);
   if (lg(y) == 1) pari_err_INV("idealdivexact", y0);
-  if (lg(x) == 1) { avma = av; return cgetg(1, t_MAT); } /* numerator is zero */
+  if (lg(x) == 1) { set_avma(av); return cgetg(1, t_MAT); } /* numerator is zero */
   y = Q_primitive_part(y, &cy);
   if (cy) x = RgM_Rg_div(x,cy);
   xZ = gcoeff(x,1,1); if (typ(xZ) != t_INT) err_divexact(x,y);
@@ -2188,7 +2440,7 @@ idealdivexact(GEN nf, GEN x0, GEN y0)
   if (typ(Nx) != t_INT) err_divexact(x,y);
   q = dvmdii(Nx,Ny, &r);
   if (signe(r)) err_divexact(x,y);
-  if (is_pm1(q)) { avma = av; return matid(nf_get_degree(nf)); }
+  if (is_pm1(q)) { set_avma(av); return matid(nf_get_degree(nf)); }
   /* Find a norm Nz | Ny such that gcd(Nx/Nz, Nz) = 1 */
   for (Nz = Ny;;) /* q = Nx/Nz */
   {
@@ -2222,7 +2474,7 @@ idealintersect(GEN nf, GEN x, GEN y)
   nf = checknf(nf);
   x = idealhnf_shallow(nf,x);
   y = idealhnf_shallow(nf,y);
-  if (lg(x) == 1 || lg(y) == 1) { avma = av; return cgetg(1,t_MAT); }
+  if (lg(x) == 1 || lg(y) == 1) { set_avma(av); return cgetg(1,t_MAT); }
   x = Q_remove_denom(x, &dx);
   y = Q_remove_denom(y, &dy);
   if (dx) y = ZM_Z_mul(y, dx);
@@ -2315,13 +2567,10 @@ RM_round_maxrank(GEN G0)
 {
   long e, r = lg(G0)-1;
   pari_sp av = avma;
-  GEN G = G0;
-  for (e = 4; ; e <<= 1)
+  for (e = 4; ; e <<= 1, set_avma(av))
   {
-    GEN H = ground(G);
+    GEN G = gmul2n(G0, e), H = ground(G);
     if (ZM_rank(H) == r) return H; /* maximal rank ? */
-    avma = av;
-    G = gmul2n(G0, e);
   }
 }
 
@@ -2340,7 +2589,7 @@ idealred0(GEN nf, GEN I, GEN vdir)
   {
     case id_PRIME:
       if (pr_is_inert(I)) {
-        if (!aI) { avma = av; return matid(N); }
+        if (!aI) { set_avma(av); return matid(N); }
         c1 = gel(I,1); I = matid(N);
         goto END;
       }
@@ -2353,7 +2602,7 @@ idealred0(GEN nf, GEN I, GEN vdir)
       IZ = gcoeff(I,1,1);
       if (is_pm1(IZ))
       {
-        if (!aI) { avma = av; return matid(N); }
+        if (!aI) { set_avma(av); return matid(N); }
         goto END;
       }
       J = idealHNF_inv_Z(nf, I);
@@ -2365,7 +2614,7 @@ idealred0(GEN nf, GEN I, GEN vdir)
   }
   /* now I integral, HNF; and J = (I\cap Z) I^(-1), integral */
   y = idealpseudomin(J, G); /* small elt in (I\cap Z)I^(-1), integral */
-  if (ZV_isscalar(y))
+  if (equalii(ZV_content(y), IZ))
   { /* already reduced */
     if (!aI) return gerepilecopy(av, I);
     goto END;
@@ -2544,6 +2793,28 @@ u_ppo(ulong x, ulong f)
   return x;
 }
 
+/* result known to be representable as an ulong */
+static ulong
+lcmuu(ulong a, ulong b) { ulong d = ugcd(a,b); return (a/d) * b; }
+
+/* assume 0 < x < N; return u in (Z/NZ)^* such that u x = gcd(x,N) (mod N);
+ * set *pd = gcd(x,N) */
+ulong
+Fl_invgen(ulong x, ulong N, ulong *pd)
+{
+  ulong d, d0, e, v, v1;
+  long s;
+  *pd = d = xgcduu(N, x, 0, &v, &v1, &s);
+  if (s > 0) v = N - v;
+  if (d == 1) return v;
+  /* vx = gcd(x,N) (mod N), v coprime to N/d but need not be coprime to N */
+  e = N / d;
+  d0 = u_ppo(d, e); /* d = d0 d1, d0 coprime to N/d, rad(d1) | N/d */
+  if (d0 == 1) return v;
+  e = lcmuu(e, d / d0);
+  return u_chinese_coprime(v, 1, e, d0, e*d0);
+}
+
 /* x t_INT, f ideal. Write x = x1 x2, sqf(x1) | f, (x2,f) = 1. Return x2 */
 static GEN
 nf_coprime_part(GEN nf, GEN x, GEN listpr)
@@ -2798,7 +3069,7 @@ idealtwoelt2(GEN nf, GEN x, GEN a)
   if (lg(x) == 1)
   {
     if (!isintzero(a)) not_in_ideal(a);
-    avma = av; return gen_0;
+    set_avma(av); return gen_0;
   }
   x = Q_primitive_part(x, &cx);
   if (cx) a = gdiv(a, cx);
