@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -492,11 +493,10 @@ lllintpartial_inplace(GEN mat) { return lllintpartialall(mat,0); }
 /********************************************************************/
 
 static int
-check(double beta, double tau, double rho, long d, long delta, long t)
+check(double b, double x, double rho, long d, long dim, long delta, long t)
 {
-  long dim = d*delta + t;
-  double cond = delta * (d * (delta+1) - 2*beta*dim + rho * (delta-1 + 2*t))
-                + tau*dim*(dim - 1);
+  double cond = delta * (d * (delta+1) - 2*b*dim + rho * (delta-1 + 2*t))
+                + x*dim*(dim - 1);
   if (DEBUGLEVEL >= 4)
     err_printf("delta = %d, t = %d (%.1lf)\n", delta, t, cond);
   return (cond <= 0);
@@ -505,25 +505,21 @@ check(double beta, double tau, double rho, long d, long delta, long t)
 static void
 choose_params(GEN P, GEN N, GEN X, GEN B, long *pdelta, long *pt)
 {
-  long d = degpol(P);
+  long d = degpol(P), dim;
   GEN P0 = leading_coeff(P);
-  double logN = gtodouble(glog(N, DEFAULTPREC)), tau, beta, rho;
-  long delta, t;
-  tau = gtodouble(glog(X, DEFAULTPREC)) / logN;
-  beta = B? gtodouble(glog(B, DEFAULTPREC)) / logN: 1.;
-  if (tau * d >= beta * beta)
-    pari_err_OVERFLOW("zncoppersmith [bound too large]");
-  /* TODO : remove P0 completely ! */
-  rho = gtodouble(glog(P0, DEFAULTPREC)) / logN;
+  double logN = gtodouble(glog(N, DEFAULTPREC)), x, b, rho;
+  x = gtodouble(glog(X, DEFAULTPREC)) / logN;
+  b = B? gtodouble(glog(B, DEFAULTPREC)) / logN: 1.;
+  if (x * d >= b * b) pari_err_OVERFLOW("zncoppersmith [bound too large]");
+  /* TODO : remove P0 completely */
+  rho = is_pm1(P0)? 0: gtodouble(glog(P0, DEFAULTPREC)) / logN;
 
-  /* Enumerate (delta,t) by increasing dimension of resulting lattice.
-   * Not subtle, but o(1) for computing time */
-  t = d; delta = 0;
-  for(;;)
+  /* Enumerate (delta,t) by increasing lattice dimension */
+  for(dim = d + 1;; dim++)
   {
-    t += d * delta + 1; delta = 0;
-    for (delta = 0; t >= 0; delta++, t -= d)
-      if (check(beta,tau,rho,d,delta,t)) { *pdelta = delta; *pt = t; return; }
+    long delta, t; /* dim = d*delta + t in the loop */
+    for (delta = 0, t = dim; t >= 0; delta++, t -= d)
+      if (check(b,x,rho,d,dim,delta,t)) { *pdelta = delta; *pt = t; return; }
   }
 }
 
@@ -539,7 +535,7 @@ do_exhaustive(GEN P, GEN N, long x, GEN B)
   long j;
   RgX_even_odd(P, &Pe,&Po); av = avma;
   if (sol_OK(gel(P,2), N,B)) vecsmalltrunc_append(sol, 0);
-  for (j = 1; j <= x; j++, avma = av)
+  for (j = 1; j <= x; j++, set_avma(av))
   {
     GEN j2 = sqru(j), E = FpX_eval(Pe,j2,N), O = FpX_eval(Po,j2,N);
     if (sol_OK(addmuliu(E,O,j), N,B)) vecsmalltrunc_append(sol, j);
@@ -554,7 +550,7 @@ GEN
 zncoppersmith(GEN P, GEN N, GEN X, GEN B)
 {
   GEN Q, R, N0, M, sh, short_pol, *Xpowers, sol, nsp, cP, Z;
-  long delta, i, j, row, d, l, dim, t, bnd = 10;
+  long delta, i, j, row, d, l, t, dim, bnd = 10;
   const ulong X_SMALL = 1000;
   pari_sp av = avma;
 
@@ -566,7 +562,7 @@ zncoppersmith(GEN P, GEN N, GEN X, GEN B)
   }
   if (signe(X) < 0) pari_err_DOMAIN("zncoppersmith", "X", "<", gen_0, X);
   P = FpX_red(P, N); d = degpol(P);
-  if (d == 0) { avma = av; return cgetg(1, t_VEC); }
+  if (d == 0) { set_avma(av); return cgetg(1, t_VEC); }
   if (d < 0) pari_err_ROOTS0("zncoppersmith");
   if (B && typ(B) != t_INT) B = gceil(B);
   if (abscmpiu(X, X_SMALL) <= 0)
@@ -594,7 +590,6 @@ zncoppersmith(GEN P, GEN N, GEN X, GEN B)
   for(;;)
   {
     dim = d * delta + t;
-
     /* TODO: In case of failure do not recompute the full vector */
     Xpowers = (GEN*)new_chunk(dim + 1);
     Xpowers[0] = gen_1;
@@ -662,7 +657,7 @@ zncoppersmith(GEN P, GEN N, GEN X, GEN B)
     if (cmpii(nsp, mului(bnd, Z)) < 0) break; /* SUCCESS */
 
     /* Failed with the precomputed or supplied value */
-    t++; if (t == d) { delta++; t = 1; }
+    if (++t == d) { delta++; t = 1; }
     if (DEBUGLEVEL >= 2)
       err_printf("Increasing dim, delta = %d t = %d\n", delta, t);
   }
@@ -738,7 +733,7 @@ lindep_bit(GEN x, long bit)
 {
   pari_sp av = avma;
   GEN v, M = lindepfull_bit(x,bit);
-  if (!M) { avma = av; return cgetg(1, t_COL); }
+  if (!M) { set_avma(av); return cgetg(1, t_COL); }
   v = gel(M,1); setlg(v, lg(M));
   return gerepilecopy(av, v);
 }
@@ -938,7 +933,7 @@ seralgdep(GEN s, long p, long r)
       gel(v, r*n + m + 1) = c;
     }
   D = lindep_Xadic(v);
-  if (lg(D) == 1) { avma = av; return gen_0; }
+  if (lg(D) == 1) { set_avma(av); return gen_0; }
   v = cgetg(p+1, t_VEC);
   for (n = 0; n < p; n++)
     gel(v, n+1) = RgV_to_RgX(vecslice(D, r*n+1, r*n+r), vy);
@@ -1068,6 +1063,14 @@ ZM_zc_mul_canon(GEN u, GEN x)
   return ZC_canon(ZM_zc_mul(u,x));
 }
 
+static GEN
+ZM_zc_mul_canon_zm(GEN u, GEN x)
+{
+  pari_sp av = avma;
+  GEN M = ZV_to_zv(ZC_canon(ZM_zc_mul(u,x)));
+  return gerepileupto(av, M);
+}
+
 struct qfvec
 {
   GEN a, r, u;
@@ -1126,6 +1129,7 @@ forqfvec_i(void *E, long (*fun)(void *, GEN, GEN, double), struct qfvec *qv, GEN
   {
     BORNE = gfloor(BORNE);
     if (typ(BORNE) != t_INT) pari_err_TYPE("minim0",BORNE);
+    if (signe(BORNE) <= 0) return;
   }
   if (n == 1) return;
   minim_alloc(n, &q, &x, &y, &z, &v);
@@ -1185,31 +1189,39 @@ forqfvec(void *E, long (*fun)(void *, GEN, GEN, double), GEN a, GEN BORNE)
   struct qfvec qv;
   forqfvec_init(&qv, a);
   forqfvec_i(E, fun, &qv, BORNE);
-  avma = av;
+  set_avma(av);
 }
 
+struct qfvecwrap
+{
+  void *E;
+  long (*fun)(void *, GEN);
+};
+
 static long
-_gp_forqf(void *E, GEN u, GEN x, double p/*unused*/)
+forqfvec_wrap(void *E, GEN u, GEN x, double d)
 {
   pari_sp av = avma;
-  (void)p;
-  set_lex(-1, ZM_zc_mul_canon(u, x));
-  closure_evalvoid((GEN)E);
-  avma = av;
-  return loop_break();
+  struct qfvecwrap *W = (struct qfvecwrap *) E;
+  (void) d;
+  return gc_long(av, W->fun(W->E, ZM_zc_mul_canon(u, x)));
+}
+
+void
+forqfvec1(void *E, long (*fun)(void *, GEN), GEN a, GEN BORNE)
+{
+  pari_sp av = avma;
+  struct qfvecwrap wr;
+  struct qfvec qv;
+  wr.E = E; wr.fun = fun;
+  forqfvec_init(&qv, a);
+  forqfvec_i((void*) &wr, forqfvec_wrap, &qv, BORNE);
+  set_avma(av);
 }
 
 void
 forqfvec0(GEN a, GEN BORNE, GEN code)
-{
-  pari_sp av = avma;
-  struct qfvec qv;
-  forqfvec_init(&qv, a);
-  push_lex(gen_0, code);
-  forqfvec_i((void*) code, &_gp_forqf, &qv, BORNE);
-  pop_lex(1);
-  avma = av;
-}
+{ EXPRVOID_WRAP(code, forqfvec1(EXPR_ARGVOID, a,  BORNE)) }
 
 enum { min_ALL = 0, min_FIRST, min_VECSMALL, min_VECSMALL2 };
 
@@ -1219,7 +1231,7 @@ enum { min_ALL = 0, min_FIRST, min_VECSMALL, min_VECSMALL2 };
  *   u[2]= maximum norm found
  *   u[3]= list of vectors found (at most STOCKMAX, unless NULL)
  *
- *  If BORNE = NULL: Minimal non-zero vectors.
+ *  If BORNE = NULL: Minimal nonzero vectors.
  *  flag = min_ALL,   as above
  *  flag = min_FIRST, exits when first suitable vector is found.
  *  flag = min_VECSMALL, return a t_VECSMALL of (half) the number of vectors
@@ -1244,7 +1256,8 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
     BORNE = gfloor(BORNE);
     if (typ(BORNE) != t_INT) pari_err_TYPE("minim0",BORNE);
     if (is_bigint(BORNE)) pari_err_PREC( "qfminim");
-    sBORNE = itos(BORNE); avma = av;
+    sBORNE = itos(BORNE); set_avma(av);
+    if (sBORNE < 0) sBORNE = 0;
   }
   if (!STOCKMAX)
   {
@@ -1348,10 +1361,10 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
     { /* maxnorm < 0 : only look for minimal vectors */
       pari_sp av2 = avma;
       gnorme = roundr(dbltor(p));
-      if (cmpis(gnorme, sBORNE) >= 0) avma = av2;
+      if (cmpis(gnorme, sBORNE) >= 0) set_avma(av2);
       else
       {
-        sBORNE = itos(gnorme); avma = av1;
+        sBORNE = itos(gnorme); set_avma(av1);
         BOUND = sBORNE * (1+eps);
         L = new_chunk(maxrank+1);
         s = 0;
@@ -1389,16 +1402,18 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
   switch(flag)
   {
     case min_FIRST:
-      avma = av; return cgetg(1,t_VEC);
+      set_avma(av); return cgetg(1,t_VEC);
     case min_VECSMALL:
     case min_VECSMALL2:
-      avma = (pari_sp)L; return L;
+      set_avma((pari_sp)L); return L;
   }
   r = (maxnorm >= 0) ? roundr(dbltor(maxnorm)): stoi(sBORNE);
   k = minss(s,maxrank);
   L[0] = evaltyp(t_MAT) | evallg(k + 1);
   if (dolll)
-    for (j=1; j<=k; j++) gel(L,j) = ZM_zc_mul_canon(u, gel(L,j));
+    for (j=1; j<=k; j++)
+      gel(L,j) = dolll==1 ? ZM_zc_mul_canon(u, gel(L,j))
+                          : ZM_zc_mul_canon_zm(u, gel(L,j));
   return gerepilecopy(av, mkvec3(stoi(s<<1), r, L));
 }
 
@@ -1410,6 +1425,14 @@ minim0(GEN a, GEN BORNE, GEN STOCKMAX, long flag)
   return v;
 }
 
+static GEN
+minim0_zm(GEN a, GEN BORNE, GEN STOCKMAX, long flag)
+{
+  GEN v = minim0_dolll(a, BORNE, STOCKMAX, flag, 2);
+  if (!v) pari_err_PREC("qfminim");
+  return v;
+}
+
 GEN
 qfrep0(GEN a, GEN borne, long flag)
 { return minim0(a, borne, gen_0, (flag & 1)? min_VECSMALL2: min_VECSMALL); }
@@ -1442,6 +1465,10 @@ GEN
 minim(GEN a, GEN borne, GEN stockmax)
 { return minim0(a,borne,stockmax,min_ALL); }
 
+GEN
+minim_zm(GEN a, GEN borne, GEN stockmax)
+{ return minim0_zm(a,borne,stockmax,min_ALL); }
+
 GEN
 minim_raw(GEN a, GEN BORNE, GEN STOCKMAX)
 { return minim0_dolll(a, BORNE, STOCKMAX, min_ALL, 0); }
@@ -1485,7 +1512,7 @@ addcolumntomatrix(GEN V, GEN invp, GEN L)
 }
 
 GEN
-perf(GEN a)
+qfperfection(GEN a)
 {
   pari_sp av = avma;
   GEN u, L;
@@ -1500,8 +1527,8 @@ perf(GEN a)
   {
     GEN D, V, invp;
     L = gel(L, 3); l = lg(L);
-    if (l == 2) { avma = av; return gen_1; }
-
+    if (l == 2) { set_avma(av); return gen_1; }
+    /* |L[i]|^2 fits  into a long for all i */
     D = zero_zv(r);
     V = cgetg(r+1, t_VECSMALL);
     invp = matid(r);
@@ -1513,7 +1540,7 @@ perf(GEN a)
       long i, j, I;
       for (i = I = 1; i<=n; i++)
         for (j=i; j<=n; j++,I++) V[I] = x[i]*x[j];
-      if (!addcolumntomatrix(V,invp,D)) avma = av2;
+      if (!addcolumntomatrix(V,invp,D)) set_avma(av2);
       else if (++s == r) break;
     }
   }
@@ -1523,7 +1550,7 @@ perf(GEN a)
     L = fincke_pohst(a,NULL,-1, DEFAULTPREC, NULL);
     if (!L) pari_err_PREC("qfminim");
     L = gel(L, 3); l = lg(L);
-    if (l == 2) { avma = av; return gen_1; }
+    if (l == 2) { set_avma(av); return gen_1; }
     M = cgetg(l, t_MAT);
     for (k = 1; k < l; k++)
     {
@@ -1535,7 +1562,7 @@ perf(GEN a)
     }
     s = ZM_rank(M);
   }
- avma = av; return utoipos(s);
+ set_avma(av); return utoipos(s);
 }
 
 static GEN
@@ -1550,7 +1577,7 @@ clonefill(GEN S, long s, long t)
 
 /* increment ZV x, by incrementing cell of index k. Initial value x0[k] was
  * chosen to minimize qf(x) for given x0[1..k-1] and x0[k+1,..] = 0
- * The last non-zero entry must be positive and goes through x0[k]+1,2,3,...
+ * The last nonzero entry must be positive and goes through x0[k]+1,2,3,...
  * Others entries go through: x0[k]+1,-1,2,-2,...*/
 INLINE void
 step(GEN x, GEN y, GEN inc, long k)
@@ -1574,7 +1601,7 @@ mplessthan(GEN x, GEN y)
 {
   pari_sp av = avma;
   GEN z = mpsub(x, y);
-  avma = av;
+  set_avma(av);
   if (typ(z) == t_INT) return (signe(z) < 0);
   if (signe(z) >= 0) return 0;
   if (realprec(z) > LOWDEFAULTPREC) return 1;
@@ -1588,7 +1615,7 @@ mpgreaterthan(GEN x, GEN y)
 {
   pari_sp av = avma;
   GEN z = mpsub(x, y);
-  avma = av;
+  set_avma(av);
   if (typ(z) == t_INT) return (signe(z) > 0);
   if (signe(z) <= 0) return 0;
   if (realprec(z) > LOWDEFAULTPREC) return 1;
@@ -1629,7 +1656,7 @@ check_bound(GEN B, GEN xk, GEN yk, GEN zk, GEN vk)
 {
   pari_sp av = avma;
   int f = mpgreaterthan(norm_aux(xk,yk,zk,vk), B);
-  avma = av; return !f;
+  return gc_bool(av, !f);
 }
 
 /* q(k-th canonical basis vector), where q is given in Cholesky form
@@ -1682,10 +1709,10 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
   if (BORNE)
   {
     borne1 = BORNE;
+    if (gsigne(borne1) <= 0) retmkvec3(gen_0, gen_0, cgetg(1,t_MAT));
     if (typ(borne1) != t_REAL)
     {
       long prec;
-      if (gequal0(borne1)) retmkvec3(gen_0, gen_0, cgetg(1,t_MAT));
       prec = nbits2prec(gexpo(borne1) + 10);
       borne1 = gtofp(borne1, maxss(prec, DEFAULTPREC));
     }
@@ -1760,7 +1787,7 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
 
     av1 = avma;
     norme1 = norm_aux(gel(x,1),gel(y,1),gel(z,1),gel(v,1));
-    if (mpgreaterthan(norme1,borne1)) { avma = av1; continue; /* main */ }
+    if (mpgreaterthan(norme1,borne1)) { set_avma(av1); continue; /* main */ }
 
     norme1 = gerepileuptoleaf(av1,norme1);
     if (check)
@@ -1811,7 +1838,7 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
       imax = i;
       for (i=imin, s=0; i < imax; i++) gel(S2,++s) = gel(S,per[i]);
       for (i = 1; i <= s; i++) gel(S,i) = gel(S2,i);
-      avma = av2;
+      set_avma(av2);
       if (s) { borne2 = mulrr(borne1, alpha); checkcnt = 0; }
       if (!stockall) continue;
       if (s > stockmax/2) stockmax <<= 1;
