@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -128,7 +129,7 @@ qflocalinvariants(GEN G, GEN P)
   for (j = 1; j < l; j++)
   {
     GEN g = gel(G,j);
-    if (typ(g) == t_QFI || typ(g) == t_QFR) gel(G,j) = gtomat(g);
+    if (is_qfb_t(typ(g))) gel(G,j) = gtomat(g);
   }
   sol = cgetg(l, t_MAT);
   if (lg(gel(G,1)) == 3)
@@ -423,7 +424,7 @@ blocks4(GEN G, long d, long n, GEN *A, GEN *B, GEN *C)
  * Returns [G',U,factd] with U in GLn(Q) such that G'=U~*G*U*constant
  * is integral and has minimal determinant.
  * In dimension 3 or 4, may return a prime p if the reduction at p is
- * impossible because of local non-solvability.
+ * impossible because of local nonsolvability.
  * P,E = factor(+/- det(G)), "prime" -1 is ignored. Destroy E. */
 static GEN qfsolvemodp(GEN G, GEN p);
 static GEN
@@ -547,7 +548,7 @@ qfminimize(GEN G, GEN P, GEN E)
     }
 
     /* Minimization was not possible so far. */
-    /* If n == 3 or 4, this proves the local non-solubility at p. */
+    /* If n == 3 or 4, this proves the local nonsolubility at p. */
     if (n == 3 || n == 4)
     {
       if (DEBUGLEVEL >= 1) err_printf(" no local solution at %Ps\n",p);
@@ -622,7 +623,7 @@ qfbsqrt(GEN D, GEN q, GEN P, GEN E)
     return qfb(D, c, shifti(negi(b),1), shifti(a,1));
 }
 
-/* \prod gen[i]^e[i] as a Qfb, e in {0,1}^n non-zero */
+/* \prod gen[i]^e[i] as a Qfb, e in {0,1}^n nonzero */
 static GEN
 qfb_factorback(GEN D, GEN gen, GEN e)
 {
@@ -676,12 +677,12 @@ quadclass2(GEN D, GEN P2D, GEN E2D, GEN Pm2D, GEN W, int n_is_4)
   if (vD == 2 && smodis(D,16) != 4)
   {
     GEN q2 = qfb(D, gen_2,gen_2, shifti(subsi(4,D),-3));
-    m++; r++; gen = shallowconcat(gen, mkvec(q2));
+    m++; r++; gen = vec_append(gen, q2);
   }
   if (vD == 3)
   {
     GEN q2 = qfb(D, gen_2,gen_0, negi(shifti(D,-3)));
-    m++; r++; gen = shallowconcat(gen, mkvec(q2));
+    m++; r++; gen = vec_append(gen, q2);
   }
   if (!r) return id(D);
   Wgen = qflocalinvariants(gen,Pm2D);
